#ifndef Submat_H
#define Submat_H
#include<iostream>
#include<iomanip>
#include<fstream>
#include<cstdio>
#include<cstdlib>
#include<cstring>
#include<cassert>
#include<cmath>
#include<string>
#include<stdint.h>
#include<map>
#include<vector>
#include<sstream>
#include<time.h>
#include "DirichletClass.h"
using namespace std ;

class SubmatClass_t {
    private:
        string alphabet;
        map<char,int> aa2idx;
        int K;
        vector<vector<double> > Mpow1; // base mutability matrix
        vector<vector<double> >     S; // matrix of eigen vectors of Mpow1
        vector<vector<double> >     D; // diagonal matrix of eigen vectors of Mpow1
        vector<vector<double> >  invS; // inverse of S of Mpow1
        vector<vector<double> > MpowK; // computed on demand, and overwritten
        vector<vector<double> > readMatrix(string);
        void                    normalizeMpowK(); // ensure MpowK each colsum=1

        /* Inferred Dirichlet priors on the algn 3 state automaton params
         * associated with K, in the range [1,1000].
         * 3 state automaton is over a SYMMETRIC Match, Insert and Delete states.
         * */
        vector<vector<DirichletClass_t> > dirichletpriors; 
        vector<vector<DirichletClass_t> > readDirichletPriors(string);
    public:
        SubmatClass_t(int n=1);
        string getAlphabet();
        void   exp(int);
        double pr(char,char); // derives probability of elemement from MpowK matrix
        double pr(unsigned int,unsigned int); // ditto, with a different interface 
        double getS(int,int);
        double getSinv(int,int);
        double getD(int,int);
        int getIndx(char);
        int getK();
        DirichletClass_t getDirichlet(string state);
        void printMpowK(); // print current state of PAM M^K matrix with stats
};
#endif
