#include "cmdLineParser.h"

bool doesFileExist(string fname) {

    const char *fn  = fname.c_str();
    ifstream infile(fn,ios::in);
    return infile.good();
}

/* Pretty LAME cmdline parser */
void parseCmdLine( 
  int nargs, 
  char *args[],
  vector<string> &seqfn,
  bool &param_ivalue,
  string &param_input_afasta,
  string &param_output_prefix,
  string &param_alignmentType 
) {
    bool continueflag = true ;
    multimap<string,string> myMap;

    string currentSwitch = "-seq";
    for( int i = 1 ; i <  nargs ; i++ ) {
        if( args[i][0]== '-' ) {
            if( strcmp(args[i],"--ivalue") == 0 ) {
                param_ivalue = true;
                currentSwitch = args[i];
            }
            else if( 
                strcmp(args[i],"-o") == 0 
                || strcmp(args[i],"--criterion") == 0 
            ) {
                currentSwitch = args[i];
            }
            else  {
                cerr << "::CmdLineError:: Unsupported option: " 
                    << args[i] << endl ;
                continueflag = false ;
                currentSwitch.assign("-X");
            }
        }
        else{
            myMap.insert( pair<string, string>(currentSwitch,args[i])) ;
        }   
    }
    if(param_ivalue == true && myMap.count("--ivalue") != 1) {
        cerr << "::CmdLineError:: Option \'--ivalue\' takes an alignment fasta file as an argument\n" ;
        continueflag = false;
    }

    if(param_ivalue == false && myMap.count("-seq") != 2) {
        cerr << "::CmdLineError:: " << "Expecting two sequences (in fasta format) as arguments to align, or one alignment (in fasta format) to score. See usage below\n";
        continueflag = false;
    }

    if(myMap.count("-o") > 1) {
        cerr << "::CmdLineError:: Option \'-o\' takes one output_afasta_string as an argument\n";
        continueflag = false;
    }
    
    if(myMap.count("--criterion") > 1) {
        cerr << "::CmdLineError:: Option \'--criterion\' takes exactly one argument\n";
        continueflag = false;
    }


    if(myMap.count("-X") > 0 ) {
        cerr << "::CmdLineError:: The following arguments cannot be processed:\n" ;
        pair<multimap<string, string>::iterator, multimap<string, string>::iterator> range;
        range = myMap.equal_range("-X");

        /* Loop through range of maps of key "-X" */
        for (multimap<string, string>::iterator itr = range.first;
                     itr != range.second;
             ++itr) {
                cerr << "                  " << (*itr).second << endl;
        }
        continueflag = false;
    }

    //get seq fasta fnames from arguments
    if(param_ivalue == false && myMap.count("-seq") == 2 ) {
        pair<multimap<string, string>::iterator, multimap<string, string>::iterator> range;
        range = myMap.equal_range("-seq");

        /* Loop through range of maps of key "-seq" */
        multimap<string, string>::iterator itr;
        for (itr = range.first; itr != range.second; ++itr) {
            string fname = (*itr).second; 
            if (!doesFileExist(fname)) {
                cerr << "::CmdLineError:: File '" << fname << "' does not exist\n" ;
                continueflag = false;
            }
            seqfn.push_back(fname);
        }
    }

    //assign fasta fname
    if(myMap.count("--ivalue") == 1 ) {
        pair<multimap<string, string>::iterator, multimap<string, string>::iterator> range;
        range = myMap.equal_range("--ivalue");

        /* Loop through range of maps of key "--ivalue" */
        multimap<string, string>::iterator itr;
        for (itr = range.first; itr != range.second; ++itr) {
            param_input_afasta = (*itr).second;
            if (!doesFileExist(param_input_afasta)) {
                cerr << "::CmdLineError:: File '" << param_input_afasta << "' does not exist\n" ;
                continueflag = false;
            }
        }
    }

    if(myMap.count("--criterion") == 1 ) {
        pair<multimap<string, string>::iterator, multimap<string, string>::iterator> range;
        range = myMap.equal_range("--criterion");

        multimap<string, string>::iterator itr;
        for (itr = range.first; itr != range.second; ++itr) {
            param_alignmentType = (*itr).second;
            if (
                   param_alignmentType.compare("marginal") !=0 
                && param_alignmentType.compare("marginal-interactive") !=0 
                && param_alignmentType.compare("optimal") !=0
            ) {
                cerr << "::CmdLineError:: Invalid option for --criterion: " << param_alignmentType << "\n" ;
                continueflag = false;
            }
        }
    }


    if(continueflag == false || nargs < 2 ) {
        cerr  << "\nTo find/compute alignment(s) between protein sequences in the MML information framework\n";
        cerr  << "    Usage: " << "seqmmligner <seq1 fasta> <seq2 fasta>\n";
        cerr  << "           " << "           (--criterion <optimal/marginal/marginal-interactive>)\n";
        cerr  << "           " << "           ( -o  <output_prefix_string>) \n";
        cerr  << "To score/evaluate any pre-computed alignments in the  MML information framework\n";
        cerr  << "    Usage: " << "seqmmligner --ivalue <alignment fasta>)\n";
        exit(1) ;
    }
    if (param_ivalue == false) assert(seqfn.size() == 2);

    //assign output_prefix_string
    if(myMap.count("-o") == 1 ) {
        pair<multimap<string, string>::iterator, multimap<string, string>::iterator> range;
        range = myMap.equal_range("-o");

        /* Loop through range of maps of key "-s" */
        multimap<string, string>::iterator itr;
        for (itr = range.first; itr != range.second; ++itr) {
            param_output_prefix = (*itr).second;
        }
    }
    else {
        string outPrefixString = "output";
        /*
        outPrefixString.append(getFilenameFromPath(seqfn[0].c_str()));
        outPrefixString += "_vs_";
        outPrefixString.append(getFilenameFromPath(seqfn[1].c_str()));
        */
        param_output_prefix = outPrefixString;
    }

}
