#include<iostream>
#include<fstream>
#include<cstdio>
#include<cstdlib>
#include<cstring>
#include<cassert>
#include<string>
#include<stdint.h>
#include<limits>
#include<map>
#include<vector>
#include<sstream>
#include<time.h>
#include <algorithm>

#include "misc.h"
#include "SubmatClass.h"
#include "nullModelClass.h"
#include "threeStateDPAClass.h"
#include "cmdLineParser.h"
#define VERSIONSTRING "1.2"

using namespace std;


void coutHeader() {
  cout << "\033[0;33;1;44m" << "   Seq" << "\033[0;0m";
  cout << "\033[0;37;1;44m" << "MMLigner (v" 
      << "\033[0;31;1;40m"
      << VERSIONSTRING 
      << "\033[0;37;1;44m"
      << ") " << "\033[0;0m";
  cout << "\033[0;37;1;44m";
  cout << setw(52) 
        << "Minimum Message Length (MML) based aligner of protein "
       << "\033[0;0m";
  cout << "\033[0;33;1;44m" << "SEQUENCES   " << "\033[0;0m\n";
  cout << "\033[0;1;37;1;45m" ;
  cout << "    Authors:" ; 
  cout << "\033[0;1;37;1;45m" ;
  cout << "    Dinithi N. Sumanaweera, Lloyd Allison, and Arun S. Konagurthu.          \033[0m\n";
  cout << "\033[1;37;1;45m";
  cout << "    Reference:" ; 
  cout << "\033[0;1;37;1;45m" ;
  cout << "  Bioinformatics, 35(14): i360-i369 2019 (ISMB/ECCB 2019 Proceedings)     \033[0m\n";
  cout << left << setw(82);
  cout << endl;
}


vector<string> readFastaFile(char fname[]) {
  vector<string> seqs;

  ifstream infile(fname,ios::in);
  assert(infile);
  string astr;
  char buff[10000];
  size_t cntr = 0 ;
  while (!infile.eof()) {
    infile.getline(buff,10000);
    if (infile.eof()) {
        seqs.push_back(astr);
        break;
    }
    if (buff[0] == '>')  {
      if (cntr > 0) {
        seqs.push_back(astr);
      }
      cntr++;
      astr = "";
    }
    else {
      astr.append(buff);
    }
  }
  infile.close();
 return seqs;
}

vector<string> getAASeqs(vector<string> fn) { //expects files in fasta format
    vector<string> s1 = readFastaFile((char *)fn[0].c_str());
    vector<string> s2 = readFastaFile((char *)fn[1].c_str());
    assert(s1.size() == 1);
    assert(s2.size() == 1);
    int n = s1.size()+s2.size();
    if (n != 2) {
        cerr << "Expecting 2 amino acid sequences. Found " << n << "instead. Fix!\n";
        exit(1);
    }
    vector<string> s;
    s.push_back(s1[0]);
    s.push_back(s2[0]);
    return s;
}

/* Returns length of encoding the characters in the passed string
 * using the passed null model. Does   NOT   include the code len
 * of encoding the string length*/
double getNullEncodingLength(
 string s, 
 nullModelClass_t &nullMdl
) {
    double cdlen = 0;
    size_t order = nullMdl.getOrder();
    size_t alphabetSize = nullMdl.getAlphabetSize();

    //Encode symbols without full context uniformly
    //over the alphabet
    double p;
    for (int i = 0; i < order; i++) {
        p = (double)1/alphabetSize;
        cdlen += -log(p); 
        //cout << p << " " << nits2bits(-log(p)) << " " << nits2bits(cdlen) << endl;
    }
    for (int i = order; i < s.length(); i++) {
        p = nullMdl.pr(s.substr(i-order,order+1));
        cdlen += -log(p);
        //cout << s.substr(i-order,order+1) << " --> " 
        //     << p << " " << nits2bits(-log(p)) << " " << nits2bits(cdlen) << endl;
    }
    return cdlen;
}

double computeINullST(
 string &S,
 string &T,
 nullModelClass_t &nullMdl) {

  size_t nS = S.size();
  size_t nT = T.size();


  double nullModelMsgLen_S = wtCodeLen_nits(nS) + getNullEncodingLength(S, nullMdl);
  double nullModelMsgLen_T = wtCodeLen_nits(nT) + getNullEncodingLength(T, nullMdl);

  cout << "-------------------" << getTerminalColorString("NULL MODEL STATS (SEQUENCE)",45) <<"--------------------\n";
  cout << "NULL(S)          = " << setw(10) << setprecision(3) << fixed << nits2bits(nullModelMsgLen_S) << " bits ";
  cout << "(" << setw(5) << setprecision(3) << fixed << nits2bits(nullModelMsgLen_S)/nS << " bits-per-aa " ;
  cout << "over " << nS << " a.a.)" << endl;
  cout << "NULL(T)          = " << setw(10) << setprecision(3) << fixed << nits2bits(nullModelMsgLen_T) << " bits ";
  cout << "(" << setw(5) << setprecision(3) << fixed << nits2bits(nullModelMsgLen_T)/nT << " bits-per-aa " ;
  cout << "over " << nT << " a.a.)" << endl;
  stringstream ss;
  ss << setprecision(3) << fixed << nits2bits(nullModelMsgLen_S)+nits2bits(nullModelMsgLen_T);
  cout << "NULL(<S,T>)      =" << getTerminalColorString(ss.str(),43) << "bits ";
  cout << "(" << setw(5) << setprecision(3) << fixed << (nits2bits(nullModelMsgLen_S)+nits2bits(nullModelMsgLen_T))
                 /(double)(nS+nT) << " bits-per-aa " ;
  cout << "over " << nS+nT << " a.a.)" << endl << endl;
  cout << "--------------------------------------------------------------------\n";
  cout << endl;

  return nullModelMsgLen_S+nullModelMsgLen_T;
}

string removeNonStandardAA(string s) {
    string sanitized_s;
    for (size_t i = 0; i < s.length(); i++) {
        switch (s[i]) {
            case 'A':
            case 'C':
            case 'D':
            case 'E':
            case 'F':
            case 'G':
            case 'H':
            case 'I':
            case 'K':
            case 'L':
            case 'M':
            case 'N':
            case 'P':
            case 'Q':
            case 'R':
            case 'S':
            case 'T':
            case 'V':
            case 'W':
            case 'Y': sanitized_s.push_back(s[i]);
                      break;
            default: //cout << "ignoring " << s[i] << " at pos = " << i+1 << endl;
                     break;
        }
    }
    return sanitized_s;
}

vector<string> removeNonStandardAAColumnsInAlgn(vector<string> a) {
    vector<string> sanitized_a(2);
    for (size_t i = 0; i < a[0].length(); i++) {
        string algncolchars = "";
        algncolchars += a[0][i];
        algncolchars += a[1][i];
        assert(algncolchars.length()==2);
        bool ignoreAlgnColFlag = false;
        for (size_t j = 0; j < 2; j++) {
            switch (algncolchars[j]) {
                case '-':
                case 'A':
                case 'C':
                case 'D':
                case 'E':
                case 'F':
                case 'G':
                case 'H':
                case 'I':
                case 'K':
                case 'L':
                case 'M':
                case 'N':
                case 'P':
                case 'Q':
                case 'R':
                case 'S':
                case 'T':
                case 'V':
                case 'W':
                case 'Y': break;
                default: ignoreAlgnColFlag = true;
                         cout << algncolchars << endl;
                         break;
            }
        }
        if (ignoreAlgnColFlag == false) {
            if (algncolchars[0] != '-' || algncolchars[1] != '-') { //...this ignores (-,-) columns
                sanitized_a[0].push_back(algncolchars[0]);
                sanitized_a[1].push_back(algncolchars[1]);
            }
        }
    }
    return sanitized_a;
}


int main( int argc , char *argv[] ) {
  coutHeader();
  vector<string> param_seqfn;
  bool param_ivalue = false;
  string param_input_afasta;
  string param_output_prefix;
  string param_priorType = "dirichlet-mode";
  string param_nullMarkovOrder = "0";
  string param_matchStatementType = "symmetric";
  int    param_startingSubmat_K = 250;
  string param_alignmentType = "optimal"; 
  bool param_interactive = false;

  parseCmdLine(argc,argv,param_seqfn,param_ivalue,param_input_afasta,
          param_output_prefix, param_alignmentType);
 

  /* get SubmatClass_t obj */
  SubmatClass_t pam(param_startingSubmat_K); // this object defines pam mutability matrix (M^1)

  /* get nullModelClass_t obj */
  nullModelClass_t nullMdl(param_nullMarkovOrder);
  assert(nullMdl.isCompatible(pam.getAlphabet()));


  // check if you are running the program in:
  //   ---> alignment mode (i.e. param_ivalue == false)
  //   or
  //   ---> ivalue computation mode (i.e. param_ivalue == true)
  if (param_ivalue == false) { // that is, run in alignment mode
      /* read sequences*/
      vector<string> seqs = getAASeqs(param_seqfn);
      string S = removeNonStandardAA(seqs[0]);
      string T = removeNonStandardAA(seqs[1]);

      cout << setw(1);
      cout << ".................................................Running in"
          << getTerminalColorString(param_alignmentType,42)  << "mode\n";
      printSeq2Term(S,"Seq S");
      printSeq2Term(T,"Seq T");

      /* Compute Null model message length */
      double nullModelMesgLen = computeINullST(S,T,nullMdl);

      /* Compute Alignment model message length */
      threeStateDPAClass_t algnobj(S,T,nullMdl,pam,
              param_alignmentType,param_output_prefix);
      double modelMsgLen;
      if (param_alignmentType.compare("optimal")==0) {
          modelMsgLen = algnobj.getOptAlignMsgLen();
      }
      else modelMsgLen = algnobj.getMarginalProbMsgLen();

      /* Compression */
      double compression = nullModelMesgLen-modelMsgLen;

      if (param_alignmentType.compare("optimal")==0) {
          //cout << "\n----------------------ALIGNMENT SUMMARY---------------------------\n";
          string algnfsastr = algnobj.getOptAlignFSAstr();
          /* write alignment fasta file */
          vector<string> algn = fsastr2Algn(algnfsastr,S,T);
          assert(algn.size()==2);
          assert(algn[0].length() == algnfsastr.length());
          assert(algn[1].length() == algnfsastr.length());
          //printFSAstr2Algn(algnfsastr,S,T);

          string outfn = param_output_prefix+".afasta";
          ofstream algnout(outfn.c_str(), ios::out);
          assert(algnout);


          algnout << ">aligned seq 1" << endl;
          for (size_t i = 0; i < algnfsastr.length(); i++) {
              algnout << algn[0][i];
              if ((i+1)%60 == 0 || (i+1) == algnfsastr.length()) algnout << endl;
          }
          algnout << endl;

          algnout << ">aligned seq 2" << endl;
          for (size_t i = 0; i < algnfsastr.length(); i++) {
              algnout << algn[1][i];
              if ((i+1)%60 == 0 || (i+1) == algnfsastr.length()) algnout << endl;
          }
          cout << "Alignment fasta (afasta) file written to:  " 
              << getTerminalColorString(param_output_prefix + ".afasta", 43) << endl;
          cout << "\n----------------------------+=o0o=+-------------------------------\n";
          cout <<  getTerminalColorString("COMPRESSION",46) << "wrt null =   ";
          stringstream ss;
          ss << nits2bits(compression);
          if (compression >0) cout << getTerminalColorString(ss.str(), 42);
          else cout << getTerminalColorString(ss.str(), 41);
          cout << " bits" << endl;
          if (compression > 0) {
            cout << " According to MML, this alignment hypothesis is 2^{"
                << nits2bits(compression) 
                << "} times" <<  getTerminalColorString("MORE LIKELY",42) << " than null.\n";
          }
          else {
            cout << "According to MML, this alignment hypothesis is worse than null, so should be"
                << getTerminalColorString("REJECTED",41) << "\n";
            cout << "Suggestion: check the marginal probability of <S,T> by running in \'"
                << getTerminalColorString("--criterion marginal",45) << "\' mode.\n";
          }
          cout << "----------------------------+=o0o=+-------------------------------\n";
      }
      else if (param_alignmentType.compare("marginal")==0) {
          cout << "\n----------------------------+=o0o=+-------------------------------\n";
          cout <<  getTerminalColorString("COMPRESSION",46) << "wrt null =   ";
          stringstream ss;
          ss << nits2bits(compression);
          if (compression >0) cout << getTerminalColorString(ss.str(), 42);
          else cout << getTerminalColorString(ss.str(), 41);
          cout << " bits" << endl;
          if (compression > 0) {
             if (nits2bits(compression) > 6){
                cout << "According to MML, the hypothesis that <S,T> are related is 2^{"
                  << nits2bits(compression) 
                  << "} times MORE LIKELY than null.\n";
             }
             else {
                 cout << "According to MML, the hypothesis that <S,T> are related is very weak but positive.\n";
             }
          }
          else {
               cout << "According to MML, the hypothesis that <S,T> are related is worse than null, so NOT related.\n";
          }
          cout << "----------------------------+=o0o=+-------------------------------\n";
      }
      else {
          cout << "----------------------------+=o0o=+-------------------------------\n";
      }
  }
  else { //if (param_ivalue == true) then run in  --ivalue mode
      cout << "\n-------------IVALUE STATISTICS OF A GIVEN ALIGNMENT---------------\n";
      vector<string> algn = readFastaFile((char*)param_input_afasta.c_str());
      assert (algn.size() == 2);
      assert (algn[0].size() == algn[1].size());
      vector<string> sanitized_algn = removeNonStandardAAColumnsInAlgn(algn);
      string algnfsastr = algn2Fsastr(sanitized_algn);
      string S = removeNonStandardAA(sanitized_algn[0]);
      string T = removeNonStandardAA(sanitized_algn[1]);
      printSeq2Term(sanitized_algn[0],"Alignment String of S");
      printSeq2Term(sanitized_algn[1],"Alignment String of T");


      /* Compute Null model message length */
      double nullModelMesgLen = computeINullST(S,T,nullMdl);

      /* prepare ivalue stats*/
      threeStateDPAClass_t ivalueobj(S,T,algnfsastr,nullMdl,pam);

      double algnModelMsgLen = ivalueobj.getOptAlignMsgLen();

      /* Compression */
      double compression = nullModelMesgLen-algnModelMsgLen;
      cout << "\n----------------------------+=o0o=+-------------------------------\n";
      cout <<  "COMPRESSION wrt null =   " << nits2bits(compression)  << " bits" << endl;
      cout << "----------------------------+=o0o=+-------------------------------\n";
  }
  return 0;
}

